//
//  BackendLocator.m
//  MyMote
//
//  Copyright 2008 Chase Douglas
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  version 2 as published by the Free Software Foundation
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#import <arpa/inet.h>
#import <netdb.h>
#import <netinet/in.h>
#import <sys/socket.h>

#import "Backend.h"
#import "BackendLocator.h"

@implementation mDNSRequest

@synthesize service;
@synthesize socket;
@synthesize runLoopSource;

@end

void receiveCallBack(CFSocketRef s, CFSocketCallBackType type, CFDataRef address, const void *data, void *info);
void resolutionCallBack(CFHostRef host, CFHostInfoType type, const CFStreamError *err, void *backends);
void DNSServiceQueryRecordCallback(DNSServiceRef service, DNSServiceFlags flags, uint32_t interfaceIndex, DNSServiceErrorType errorCode, const char *fullname, uint16_t rrtype, uint16_t rrclass, uint16_t rdlen, const void *rdata, uint32_t ttl, void *context);
void DNSServiceSocketCallback(CFSocketRef socket, CFSocketCallBackType callbackType, CFDataRef address, const void *data, void *reverseRef);

@implementation BackendLocator

@synthesize delegate;
@synthesize mDNSRequests;

- (id)init {
	if (self = [super init]) {
		mDNSRequests = [[NSMutableArray alloc] init];
		
		CFSocketContext context;
		context.version = 0;
		context.info = self;
		context.retain = NULL;
		context.release = NULL;
		context.copyDescription = NULL;
		
		if ((socket = CFSocketCreate(kCFAllocatorDefault, 0, SOCK_DGRAM, 0, kCFSocketDataCallBack, receiveCallBack, &context)) == NULL) {
			NSLog(@"Error creating socket for detecting backends: %s", strerror(errno));
			return nil;
		}
		
		int yes = 1;
		if (setsockopt(CFSocketGetNative(socket), SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes))) {
			NSLog(@"Error setting reuse address socket option: %s", strerror(errno));
		}
		
		struct sockaddr_in saddr;
		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(1900);
		saddr.sin_addr.s_addr = INADDR_ANY;
		bzero(&(saddr.sin_zero), 8);
		
		if (CFSocketSetAddress(socket, (CFDataRef)[NSData dataWithBytes:&saddr length:sizeof(saddr)])) {
			NSLog(@"CFSocketSetAddress failed: %s", strerror(errno));
			CFSocketInvalidate(socket);
			CFRelease(socket);
			socket = NULL;
			return nil;
		}
		
		if ((runLoopSource = CFSocketCreateRunLoopSource(kCFAllocatorDefault, socket, 0)) == NULL) {
			NSLog(@"CFSocketCreateRunLoopSource failed: %s", strerror(errno));
			CFSocketInvalidate(socket);
			CFRelease(socket);
			socket = NULL;
			return nil;
		}
		
		saddr.sin_addr.s_addr = inet_addr("239.255.255.250");
		if ((addrData = [[NSData alloc] initWithBytes:&saddr length:sizeof(saddr)]) == nil) {
			NSLog(@"Error allocating data for multicast address: %s", strerror(errno));
			CFSocketInvalidate(socket);
			CFRelease(socket);
			socket = NULL;
			return nil;
		}
		
		CFRunLoopAddSource(CFRunLoopGetMain(), runLoopSource, kCFRunLoopDefaultMode);
	}
	return self;
}

#define UPNP_MESSAGE "M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 2\r\nST: urn:schemas-mythtv-org:device:MasterMediaServer:1\r\n\r\n"

- (void)poll {
	CFDataRef data = (CFDataRef)[NSData dataWithBytes:UPNP_MESSAGE length:(strlen(UPNP_MESSAGE) + 1)];
	if (data == NULL) {
		return;
	}
	
	CFSocketSendData(socket, (CFDataRef)addrData, data, 0);
}

void receiveCallBack(CFSocketRef s, CFSocketCallBackType type, CFDataRef address, const void *data, void *backendLocator) {
	[(BackendLocator *)backendLocator receiveData:data];
}

- (void)receiveData:(CFDataRef)data {
	if (data && [[NSString stringWithFormat:@"%s", CFDataGetBytePtr(data)] rangeOfString:@"ST: urn:schemas-mythtv-org:device:MasterMediaServer:1"].location == NSNotFound) {
		return;
	}
	
	NSScanner *scanner = [NSScanner scannerWithString:[NSString stringWithFormat:@"%s", CFDataGetBytePtr(data)]];
	[scanner scanUpToString:@"LOCATION: http://" intoString:NULL];
	if (![scanner scanString:@"LOCATION: http://" intoString:NULL]) {
		NSLog(@"Error: Could not determine location from UPNP response");
		return;
	}
	
	NSString *stringAddr;
	if (![scanner scanUpToCharactersFromSet:[NSCharacterSet characterSetWithCharactersInString:@":/ \r\n"] intoString:&stringAddr]) {
		NSLog(@"Error: Could not determine location from UPNP response");
		return;
	}
	
	NSString *stringPort;
	if (![scanner isAtEnd]) {
		[scanner setScanLocation:([scanner scanLocation] + 1)];
		
		if ([scanner isAtEnd] || ![scanner scanCharactersFromSet:[NSCharacterSet decimalDigitCharacterSet] intoString:&stringPort]) {
			stringPort = @"6544";
		}
	}
	else {
		stringPort = @"6544";
	}
	
	struct sockaddr_in saddr;
	if (inet_aton([stringAddr UTF8String], &saddr.sin_addr)) {
		[delegate foundBackend:[Backend backendWithAddress:stringAddr hostname:nil port:[stringPort intValue]]];
	
		saddr.sin_family = AF_INET;
		CFHostRef host = CFHostCreateWithAddress(kCFAllocatorDefault, (CFDataRef)[NSData dataWithBytes:&saddr length:sizeof(struct sockaddr_in)]);
		if (host == NULL) {
			NSLog(@"Error: Could not create host based on address");
			return;
		}
		
		NSMutableDictionary *info = [[NSMutableDictionary alloc] initWithObjectsAndKeys:self, @"BackendLocator", stringPort, @"stringPort", stringAddr, @"stringAddr", nil];
		
		CFHostClientContext context;
		context.version = 0;
		context.info = info;
		context.retain = NULL;
		context.release = NULL;
		context.copyDescription = NULL;
		if (CFHostSetClient(host, resolutionCallBack, &context) == FALSE) {
			NSLog(@"Error: Could not set client for hostname resolution");
			CFRelease(host);
			return;
		}
		
		CFHostScheduleWithRunLoop(host, CFRunLoopGetMain(), kCFRunLoopDefaultMode);
		
		CFStreamError err;
		if (CFHostStartInfoResolution(host, kCFHostNames, &err) == FALSE) {
			NSLog(@"Error starting hostname resolution");
			CFRelease(host);
			return;
		}
	}
	else {
		CFHostRef host = CFHostCreateWithName(kCFAllocatorDefault, (CFStringRef)stringAddr);
		if (host == NULL) {
			NSLog(@"Error: Could not create host based on hostname");
			return;
		}
		
		NSMutableDictionary *info = [[NSMutableDictionary alloc] initWithObjectsAndKeys:self, @"BackendLocator", stringPort, @"stringPort", nil];
		
		CFHostClientContext context;
		context.version = 0;
		context.info = info;
		context.retain = NULL;
		context.release = NULL;
		context.copyDescription = NULL;
		if (CFHostSetClient(host, resolutionCallBack, &context) == FALSE) {
			NSLog(@"Error: Could not set client for address resolution");
			CFRelease(host);
			return;
		}
		
		CFHostScheduleWithRunLoop(host, CFRunLoopGetMain(), kCFRunLoopDefaultMode);
		
		CFStreamError err;
		if (CFHostStartInfoResolution(host, kCFHostAddresses, &err) == FALSE) {
			NSLog(@"Error starting address resolution");
			CFRelease(host);
			UIAlertView *alert = [[UIAlertView alloc] initWithTitle:@"Resolution Error" message:@"Error starting address resolution." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil];
			[alert show];
			[alert release];
			return;
		}
	}
}

void resolutionCallBack(CFHostRef host, CFHostInfoType type, const CFStreamError *err, void *info) {
	NSMutableDictionary *infoDict = info;
	
	if (err->error) {
		CFRelease(host);
		if (type == kCFHostAddresses) {
			NSLog(@"Error resolving address");
			return;
		}
		
		NSString *stringAddr = [infoDict objectForKey:@"stringAddr"];
		uint8_t ip[4];
		if (sscanf([stringAddr UTF8String], "%hhu.%hhu.%hhu.%hhu", ip + 3, ip + 2, ip + 1, ip) != 4) {
			NSLog(@"Error resolving address");
			return;
		}
		
		DNSServiceRef reverseRef;
		DNSServiceQueryRecord(&reverseRef, kDNSServiceFlagsForceMulticast, 0, [[NSString stringWithFormat:@"%hhu.%hhu.%hhu.%hhu.in-addr.arpa", ip[0], ip[1], ip[2], ip[3]] UTF8String], kDNSServiceType_PTR, kDNSServiceClass_IN, DNSServiceQueryRecordCallback, infoDict);
		
		[infoDict setObject:[NSData dataWithBytesNoCopy:reverseRef length:sizeof(reverseRef)] forKey:@"ServiceRef"];
		
		CFSocketContext context;
		context.version = 0;
		context.info = info;
		context.retain = NULL;
		context.release = NULL;
		context.copyDescription = NULL;
		CFSocketRef socket = CFSocketCreateWithNative(kCFAllocatorDefault, DNSServiceRefSockFD(reverseRef), kCFSocketReadCallBack, DNSServiceSocketCallback, &context);
		CFRunLoopSourceRef mDNSRunLoopSource = CFSocketCreateRunLoopSource(kCFAllocatorDefault, socket, 0);
		CFRunLoopAddSource(CFRunLoopGetMain(), mDNSRunLoopSource, kCFRunLoopDefaultMode);
		
		mDNSRequest *request = [[[mDNSRequest alloc] init] autorelease];
		request.service = reverseRef;
		request.socket = socket;
		request.runLoopSource = mDNSRunLoopSource;
		[[[infoDict objectForKey:@"BackendLocator"] mDNSRequests] addObject:request];
		
		return;
	}
	
	if (type == kCFHostAddresses || type == kCFHostNames) {
		[(BackendLocator *)[infoDict objectForKey:@"BackendLocator"] resolvedHost:host port:[infoDict objectForKey:@"stringPort"]];
	}
	else {
		NSLog(@"Error: Resolution call back with type %d", type);
		CFRelease(host);
	}
	
	[infoDict release];	
}

- (void)resolvedHost:(CFHostRef)host port:(NSString *)stringPort {
	CFArrayRef addrs;
	if ((addrs = CFHostGetAddressing(host, NULL)) == NULL) {
		NSLog(@"Error: Address resolved, but no addresses available from CFHostRef");
		CFRelease(host);
		return;
	}
	
	struct sockaddr_in *addr = (struct sockaddr_in *)CFDataGetBytePtr(CFArrayGetValueAtIndex(addrs, 0));
	NSString *stringAddr = [NSString stringWithUTF8String:inet_ntoa(addr->sin_addr)];
	
	CFArrayRef names = CFHostGetNames(host, NULL);
	if (names == NULL) {
		NSLog(@"Error: Address resolved, but no hostname available from CFHostRef");
		CFRelease(host);
		return;
	}
	
	NSString *name = [NSString stringWithString:(NSString *)CFArrayGetValueAtIndex(names, 0)];
	[delegate foundBackend:[Backend backendWithAddress:stringAddr hostname:name port:[stringPort intValue]]];
	CFRelease(host);
}

void DNSServiceSocketCallback(CFSocketRef socket, CFSocketCallBackType callbackType, CFDataRef address, const void *data, void *info) {
	DNSServiceProcessResult((DNSServiceRef)[[(NSMutableDictionary *)info objectForKey:@"ServiceRef"] bytes]);
	[(NSMutableDictionary *)info release];
	CFSocketInvalidate(socket);
	CFRelease(socket);
}

void DNSServiceQueryRecordCallback(DNSServiceRef service, DNSServiceFlags flags, uint32_t interfaceIndex, DNSServiceErrorType errorCode, const char *fullname, uint16_t rrtype, uint16_t rrclass, uint16_t rdlen, const void *rdata, uint32_t ttl, void *context) {
	[[(NSMutableDictionary *)context objectForKey:@"BackendLocator"] mDNSReverseResolutionWithName:fullname data:rdata length:rdlen port:[[(NSMutableDictionary *)context objectForKey:@"stringPort"] intValue] error:errorCode service:service];
}

- (void)mDNSReverseResolutionWithName:(const char *)name data:(const char *)data length:(uint16_t)rdlen port:(unsigned short)port error:(DNSServiceErrorType)error service:(DNSServiceRef)service {
	if (error == kDNSServiceErr_NoError) {
		int index = 0;
		NSMutableString *hostname = [[NSMutableString alloc] init];
		
		while (index < rdlen && data[index] > 0 && data[index] < kDNSServiceMaxServiceName) {
			if ([hostname length]) {
				[hostname appendString:@"."];
			}
			NSString *tmp = [[NSString alloc] initWithBytesNoCopy:(char *)(data + index + 1) length:data[index] encoding:NSASCIIStringEncoding freeWhenDone:NO];
			[hostname appendString:tmp];
			[tmp release];
			index += data[index] + 1;
		}
		
		if (index != rdlen - 1 || data[index] != '\0') {
			NSLog(@"Error resolving address");
		}
		else {
			uint8_t ip[4];
			if (sscanf(name, "%hhu.%hhu.%hhu.%hhu.in-addr.arpa.", ip + 3, ip + 2, ip + 1, ip) != 4) {
				NSLog(@"Error resolving address");
			}
			else {
				NSString *stringAddr = [NSString stringWithFormat:@"%hhu.%hhu.%hhu.%hhu", ip[0], ip[1], ip[2], ip[3]];
				[delegate foundBackend:[Backend backendWithAddress:stringAddr hostname:hostname port:port]];
			}
		}
		
		[hostname release];
	}
	else {
		NSLog(@"Error resolving address");
	}
	
	mDNSRequest *request;
	for (request in mDNSRequests) {
		if (request.service == service) {
			[request retain];
			break;
		}
	}
	
	if (!request) {
		NSLog(@"Warning: No matching request found for mDNS service");
		return;
	}
	
	[mDNSRequests removeObject:request];

	CFRunLoopRemoveSource(CFRunLoopGetMain(), request.runLoopSource, kCFRunLoopDefaultMode);
	CFRelease(request.runLoopSource);
	
	[request release];
}

- (void)dealloc {
	if (socket) {
		CFSocketInvalidate(socket);
		CFRelease(socket);
	}
	if (addrData) {
		CFRelease(addrData);
	}
	if (runLoopSource) {
		CFRunLoopRemoveSource(CFRunLoopGetMain(), runLoopSource, kCFRunLoopDefaultMode);
		CFRelease(runLoopSource);
	}
	
	for (mDNSRequest *request in mDNSRequests) {
		CFRunLoopRemoveSource(CFRunLoopGetMain(), request.runLoopSource, kCFRunLoopDefaultMode);
		CFRelease(request.runLoopSource);
		CFSocketInvalidate(request.socket);
		CFRelease(request.socket);
	}
	[mDNSRequests release];
	
	[delegate release];
	[super dealloc];
}

@end
