//
//  BackendSelect.m
//  MyMote
//
//  Copyright 2008 Chase Douglas
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  version 2 as published by the Free Software Foundation
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#import <arpa/inet.h>

#import "defines.h"
#import "Backend.h"
#import "BackendEntry.h"
#import "BackendSelect.h"
#import "BackendSettings.h"
#import "FrontendSelect.h"

@implementation BackendSelect

- (id)init {
	if (self = [super initWithStyle:UITableViewStylePlain]) {
		self.title = @"Backend";
		
		lock = [[NSLock alloc] init];
		BACKEND_LOCATOR.delegate = self;
	}
	return self;
}

- (void)loadView {
	foundBackends = [[NSMutableArray alloc] init];

	[super loadView];
	
	self.navigationItem.rightBarButtonItem = [[[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemAdd target:self action:@selector(addBackend)] autorelease];
	
	self.tableView.separatorStyle = UITableViewCellSeparatorStyleSingleLine;
	self.tableView.allowsSelectionDuringEditing = YES;
}

- (void)viewWillAppear:(BOOL)animated {
	foundBackends = [[NSMutableArray alloc] init];
	[super viewWillAppear:animated];
	[BACKEND_LOCATOR poll];
	[self.tableView reloadData];
}

- (void)foundBackend:(Backend *)backend {
	BOOL backendMatch = NO;
	
	[lock lock];
	
	if (!foundBackends) {
		[lock unlock];
		return;
	}

	for (Backend *currentBackend in foundBackends) {
		if (backend.hostname && !currentBackend.hostname && [backend.address isEqualToString:currentBackend.address]) {
			currentBackend.hostname = backend.hostname;			
			[self.tableView reloadData];
		}
		if (backend.port == currentBackend.port && ([backend.address isEqualToString:currentBackend.address] || [backend.hostname isEqualToString:currentBackend.hostname])) {
			backendMatch = YES;
		}
	}
	
	if (!backendMatch) {
		[foundBackends addObject:backend];
		[self.tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:[NSIndexPath indexPathForRow:([foundBackends count] - 1) inSection:0]] withRowAnimation:UITableViewRowAnimationLeft];
	}
		
	[lock unlock];
}

- (void)addBackend {
	BackendEntry *backendEntry = [[[BackendEntry alloc] initWithTarget:self action:@selector(hostEntered:)] autorelease];
	UINavigationController *navigationController = [[[UINavigationController alloc] initWithRootViewController:backendEntry] autorelease];
	navigationController.navigationBar.barStyle = UIBarStyleBlackOpaque;
	[self presentModalViewController:navigationController animated:YES];
}

- (void)hostEntered:(NSString *)entry {	
	struct in_addr inaddr;
	
	NSString *stringAddr;
	NSScanner *scanner = [NSScanner scannerWithString:entry];
	if (![scanner scanUpToCharactersFromSet:[NSCharacterSet characterSetWithCharactersInString:@":"] intoString:&stringAddr]) {
		NSLog(@"Error: Could not create host based on hostname");
		return;
	}

	NSString *stringPort;
	if (![scanner isAtEnd]) {
		[scanner setScanLocation:([scanner scanLocation] + 1)];
	
		if ([scanner isAtEnd] || ![scanner scanCharactersFromSet:[NSCharacterSet decimalDigitCharacterSet] intoString:&stringPort]) {
			stringPort = @"6544";
		}
	}
	else {
		stringPort = @"6544";
	}
	
	Backend *backend;
	if (inet_aton([stringAddr UTF8String], &inaddr)) {
		backend = [Backend backendWithAddress:stringAddr hostname:nil port:[stringPort intValue]];
	}
	else {
		backend = [Backend backendWithAddress:nil hostname:stringAddr port:[stringPort intValue]];
	}
	
	NSMutableArray *enteredBackends = [NSMutableArray arrayWithArray:[DEFAULTS backendArrayForKey:@"Entered Backends"]];
	if (![enteredBackends containsObject:backend]) {
		[enteredBackends addObject:backend];
		[DEFAULTS setBackendArray:enteredBackends forKey:@"Entered Backends"];
		[self.tableView reloadData];
	}
}

- (void)viewWillDisappear:(BOOL)animated {
	[super viewWillDisappear:animated];
	//[backendLocator release];
	//backendLocator = nil;
	[lock lock];
	[foundBackends release];
	foundBackends = nil;
	[lock unlock];
}

- (void)dealloc {
	//[backendLocator release];
	[lock lock];
	[foundBackends release];
	[lock unlock];
	[lock release];
	[super dealloc];
}

// UITableViewDataSource

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView {
	return 2;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section {
	switch (section) {
		case 0:
			return [foundBackends count];
			
		case 1:
			return [[DEFAULTS backendArrayForKey:@"Entered Backends"] count];
	}
	
	return 0;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath {
	UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:@"backendsCell"];
	if (!cell) {
		cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:@"backendsCell"] autorelease];
	}
	
	Backend *backend;
	switch (indexPath.section) {
		case 0:
			backend = [foundBackends objectAtIndex:indexPath.row];
			break;
			
		case 1:
			backend = [[DEFAULTS backendArrayForKey:@"Entered Backends"] objectAtIndex:indexPath.row];
			break;
	}
	
	if (backend.port == 6544) {
		cell.textLabel.text = backend.hostname ? backend.hostname : backend.address;
	}
	else {
		cell.textLabel.text = [NSString stringWithFormat:@"%@:%hu", (backend.hostname ? backend.hostname : backend.address), backend.port];
	}
	
	cell.accessoryType = UITableViewCellAccessoryDetailDisclosureButton;

	return cell;
}

// UITableViewDelegate

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section {
	switch (section) {
		case 0:
			return @"Discovered Backends";
			
		case 1:
			return @"Entered Backends";
	}
	
	return @"";
}

- (UITableViewCellEditingStyle)tableView:(UITableView *)tableView editingStyleForRowAtIndexPath:(NSIndexPath *)indexPath {
	switch (indexPath.section) {
		case 0:
			return UITableViewCellEditingStyleNone;
			
		case 1:
			return UITableViewCellEditingStyleDelete;
	}
	
	return UITableViewCellEditingStyleNone;
}

- (void)tableView:(UITableView *)tableView commitEditingStyle:(UITableViewCellEditingStyle)editingStyle forRowAtIndexPath:(NSIndexPath *)indexPath {
	NSMutableArray *enteredBackends = [NSMutableArray arrayWithArray:[DEFAULTS backendArrayForKey:@"Entered Backends"]];
	[enteredBackends removeObjectAtIndex:indexPath.row];
	[DEFAULTS setBackendArray:enteredBackends forKey:@"Entered Backends"];	
	[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath] withRowAnimation:UITableViewRowAnimationLeft];
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath {
	Backend *backend = nil;
	switch (indexPath.section) {
		case 0:
			backend = [foundBackends objectAtIndex:indexPath.row];
			break;
			
		case 1:
			backend = [[DEFAULTS backendArrayForKey:@"Entered Backends"] objectAtIndex:indexPath.row];
	}
	
	[self.navigationController pushViewController:[[[FrontendSelect alloc] initWithBackend:backend] autorelease] animated:YES];
}

- (void)tableView:(UITableView *)tableView accessoryButtonTappedForRowWithIndexPath:(NSIndexPath *)indexPath {
	Backend *backend;
	
	switch (indexPath.section) {
		case 0:
			backend = [foundBackends objectAtIndex:indexPath.row];
			break;
			
		case 1:
			backend = [[DEFAULTS backendArrayForKey:@"Entered Backends"] objectAtIndex:indexPath.row];
			break;
	}
	
	[self.navigationController pushViewController:[[[BackendSettings alloc] initWithBackendAddress:[NSString stringWithFormat:@"%@:%hu", (backend.hostname ? backend.hostname : backend.address), backend.port]] autorelease] animated:YES];
}
@end
