//
//  Channel.m
//  MyMote
//
//  Copyright 2008 Chase Douglas
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  version 2 as published by the Free Software Foundation
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#import "defines.h"
#import "Channel.h"
#import "GuideResponse.h"
#import "MythXMLParser.h"
#import "Program.h"
#import "ProgramGuideData.h"

@interface Channel (Private)

- (void)getProgramsBeforeDate:(NSDate *)beforeDate afterDate:(NSDate *)afterDate;

@end


@implementation Channel

@synthesize response;
@synthesize programs;

- (id)initWithResponse:(id)newResponse attributes:(NSDictionary *)newAttributes {
	if (self = [super init]) {
		response = newResponse;
		attributes = [newAttributes retain];
		
		programs = [[NSMutableArray alloc] init];
	}
	return self;
}

ATTRIBUTE_PROPERTY(name, @"channelName");
ATTRIBUTE_PROPERTY(number, @"chanNum");
ATTRIBUTE_PROPERTY(chanId, @"chanId");
ATTRIBUTE_PROPERTY(callSign, @"callSign");

- (void)drawInRect:(CGRect)rect context:(CGContextRef)ctx {
	CGContextAddRect(ctx, rect);
	CGContextSetRGBFillColor(ctx, 1, 1, 1, 1);
	CGContextFillPath(ctx);
	
	CGContextSetRGBFillColor(ctx, 0, 0, 0, 1);
	
	CGContextScaleCTM(ctx, 1, -1);
	CGContextTranslateCTM(ctx, 0, -rect.size.height - rect.origin.y * 2);
	
	rect.size.width /= 2;

	[self.number drawCenteredInRect:rect withFont:[UIFont boldSystemFontOfSize:([UIFont systemFontSize] * SCALE)] context:ctx];
	
	rect.origin.x += rect.size.width;
	UIImage *icon;
	if ((icon = [response.programGuideData.channelIcons objectForKey:self.callSign]) != (UIImage *)[NSNull null] && icon) {
		CGContextDrawImage(ctx, rect, icon.CGImage);
	}
	else {
		[self.callSign drawCenteredFitInRect:rect withFont:[UIFont boldSystemFontOfSize:([UIFont systemFontSize] * SCALE)] context:ctx];
	}

	CGContextTranslateCTM(ctx, 0, rect.origin.y * 2 + rect.size.height);
	CGContextScaleCTM(ctx, 1, -1);
}

- (void)drawProgramsInRect:(CGRect)rect yOffset:(CGFloat)yOffset {
	NSDateComponents *offset = [[NSDateComponents alloc] init];
	offset.second = 6 * 60 * 60 * rect.origin.x / (EPG_PROGRAMS_WIDTH * SCALE);
	NSDate *startDate = [[NSCalendar currentCalendar] dateByAddingComponents:offset toDate:response.startDate options:0];
	offset.second = 6 * 60 * 60 * (rect.origin.x + rect.size.width) / (EPG_PROGRAMS_WIDTH * SCALE);
	NSDate *endDate = [[NSCalendar currentCalendar] dateByAddingComponents:offset toDate:response.startDate options:0];
	[offset release];
	
	for (Program *program in programs) {
		if (([program.startDate compare:startDate] != NSOrderedAscending && [program.startDate compare:endDate] == NSOrderedAscending)
			|| ([program.endDate compare:endDate] != NSOrderedDescending && [program.endDate compare:startDate] == NSOrderedDescending)
			|| ([program.startDate compare:startDate] == NSOrderedAscending && [program.endDate compare:endDate] == NSOrderedDescending)) {
			
			NSAutoreleasePool *drawPool = [[NSAutoreleasePool alloc] init];

			float guideStartOffset = [[NSCalendar currentCalendar] components:(NSMinuteCalendarUnit) fromDate:response.startDate toDate:program.startDate options:0].minute;
			float guideEndOffset = [[NSCalendar currentCalendar] components:(NSMinuteCalendarUnit) fromDate:response.startDate toDate:program.endDate options:0].minute;
			
			CGRect programRect = CGRectMake(0, rect.origin.y, 0, rect.size.height);
			
			if (guideStartOffset < 0) {
				programRect.origin.x = 1;
			}
			else {
				programRect.origin.x = (guideStartOffset / response.guideMinuteLength * EPG_PROGRAMS_WIDTH + 1) * SCALE;
			}
			
			if (programRect.origin.x == 1) {
				programRect.size.width = (guideEndOffset / response.guideMinuteLength * EPG_PROGRAMS_WIDTH - 2) * SCALE;
			}
			else {
				programRect.size.width = (program.programMinuteLength / response.guideMinuteLength * EPG_PROGRAMS_WIDTH - 2) * SCALE;
			}
			
			if (programRect.origin.x + programRect.size.width >= (EPG_PROGRAMS_WIDTH - 1) * SCALE) {
				programRect.size.width = (EPG_PROGRAMS_WIDTH - 1) * SCALE - programRect.origin.x;
			}
			
			if (programRect.size.width <= 0) { // A show ended just as our guide starts or vice versa
				continue;
			}
			
			programRect.origin = CGPointMake(programRect.origin.x - rect.origin.x, programRect.origin.y - yOffset);
			
			[program drawInRect:programRect];
			
			[drawPool release];
		}
		else if ([program.startDate compare:endDate] != NSOrderedAscending) {
			break;
		}
	}
}

- (void)getExtendedPrograms {
	NSDate *min = response.endDate;
	NSDate *max = response.startDate;
	for (Program *program in programs) {
		if ([program.startDate compare:min] == NSOrderedAscending) {
			min = program.startDate;
		}
		if ([program.endDate compare:max] == NSOrderedDescending) {
			max = program.endDate;
		}
	}
	
	if ([min compare:response.startDate] == NSOrderedDescending && [max compare:response.endDate] == NSOrderedAscending) {
		[self getProgramsBeforeDate:min afterDate:max];
	}
	else if ([min compare:response.startDate] == NSOrderedDescending && [max compare:response.endDate] != NSOrderedAscending) {
		[self getProgramsBeforeDate:min afterDate:nil];
	}
	else if ([min compare:response.startDate] != NSOrderedDescending && [max compare:response.endDate] == NSOrderedAscending) {
		[self getProgramsBeforeDate:nil afterDate:max];
	}
}

- (void)getProgramsBeforeDate:(NSDate *)beforeDate afterDate:(NSDate *)afterDate {
	NSDateComponents *extension = [[NSDateComponents alloc] init];
	NSDate *startDate;
	NSDate *endDate;
	if (beforeDate && afterDate) {
		extension.day = -1;
		startDate = [[NSCalendar currentCalendar] dateByAddingComponents:extension toDate:beforeDate options:0];
		extension.day = 1;
		endDate = [[NSCalendar currentCalendar] dateByAddingComponents:extension toDate:afterDate options:0];
		parseBefore = YES;
		parseAfter = YES;
	}
	else if (beforeDate) {
		extension.day = -1;
		startDate = [[NSCalendar currentCalendar] dateByAddingComponents:extension toDate:beforeDate options:0];
		endDate = beforeDate;
		parseBefore = YES;
		parseAfter = NO;
	}
	else {
		startDate = afterDate;
		extension.day = 1;
		endDate = [[NSCalendar currentCalendar] dateByAddingComponents:extension toDate:afterDate options:0];
		parseBefore = NO;
		parseAfter = YES;
	}
	[extension release];
	
	NSDateFormatter *formatter = [[NSDateFormatter alloc] init];
	formatter.dateFormat = @"yyyy-MM-dd'T'HH:mm:ss";
	
	MythXMLParser *parser = [[[MythXMLParser alloc] initWithContentsOfURL:[NSURL URLWithString:[NSString stringWithFormat:@"http://%@/Myth/GetProgramGuide?StartTime=%@&EndTime=%@&StartChanId=%@&NumOfChannels=1&Details=1", BACKEND_LOCATION, [formatter stringFromDate:startDate], [formatter stringFromDate:endDate], self.chanId]]] autorelease];
	parser.delegate = self;
	
	if ([parser parse] == NO) {
		NSLog(@"Error Retrieving Extended Programs: %@", [[parser parserError] localizedDescription]);
		NSLog(@"URL Used: http://%@/Myth/GetProgramGuide?StartTime=%@&EndTime=%@&StartChanId=%@&NumOfChannels=1&Details=1", BACKEND_LOCATION, [formatter stringFromDate:startDate], [formatter stringFromDate:endDate], self.chanId);
	}

	[formatter release];
}

- (void)dealloc {
	[extendedProgram release];
	[programs release];
	[attributes release];
	[pool release];
	[super dealloc];
}

// NSXMLParser Delegate

- (void)parser:(NSXMLParser *)parser didStartElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qualifiedName attributes:(NSDictionary *)attributesDict {
	if ([elementName isEqualToString:@"Program"]) {
		[pool release];
		pool = nil;
		
		if (parseBefore || parseAfter) {
			if (extendedProgram) {
				if (parseBefore && [extendedProgram.startDate compare:response.startDate] == NSOrderedAscending && [extendedProgram.endDate compare:response.startDate] == NSOrderedDescending) {
					[programs insertObject:extendedProgram atIndex:0];
					parseBefore = NO;
					if (parseAfter && [extendedProgram.startDate compare:response.endDate] == NSOrderedAscending && [extendedProgram.endDate compare:response.endDate] == NSOrderedDescending) {
						parseAfter = NO;
					}
				}
				else if (parseAfter && [extendedProgram.startDate compare:response.endDate] == NSOrderedAscending && [extendedProgram.endDate compare:response.endDate] == NSOrderedDescending) {
					[programs addObject:extendedProgram];
					parseAfter = NO;
				}
			}
			
			[extendedProgram release];
			if (parseBefore || parseAfter) {
				pool = [[NSAutoreleasePool alloc] init];
				extendedProgram = [[Program alloc] initWithChannel:self attributes:attributesDict];
				parser.delegate = extendedProgram;
			}
			else {
				extendedProgram = nil;
				parser.delegate = nil;
			}
		}
		else {
			pool = [[NSAutoreleasePool alloc] init];
			Program *program = [[[Program alloc] initWithChannel:self attributes:attributesDict] autorelease];
			[programs addObject:program];
			parser.delegate = program;
		}
	}
}

- (void)parser:(NSXMLParser *)parser didEndElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qualifiedName {
	static NSArray *documents = nil;

	if ([elementName isEqualToString:@"Channel"]) {
		if (extendedProgram) {
			if (parseBefore && [extendedProgram.startDate compare:response.startDate] == NSOrderedAscending && [extendedProgram.endDate compare:response.startDate] == NSOrderedDescending) {
				[programs insertObject:extendedProgram atIndex:0];
				[extendedProgram release];
				extendedProgram = nil;
			}
			else if (parseAfter && [extendedProgram.startDate compare:response.endDate] == NSOrderedAscending && [extendedProgram.endDate compare:response.endDate] == NSOrderedDescending) {
				[programs addObject:extendedProgram];
				[extendedProgram release];
				extendedProgram = nil;
			}
			else {
				[extendedProgram release];
				extendedProgram = nil;
			}
			
			parseBefore = NO;
			parseAfter = NO;
			
			parser.delegate = nil;
		}
		else {
			if (![response.programGuideData.channelIcons objectForKey:self.callSign]) {
				if (!documents) {
					documents = [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) retain];
				}
				
				UIImage *icon = [UIImage imageWithContentsOfFile:[NSString stringWithFormat:@"%@/%@.png", [documents objectAtIndex:0], self.callSign]];
				if (icon && icon.size.height == (int)((EPG_ROW_HEIGHT - 2) * SCALE)) {
					[response.programGuideData.channelIcons setObject:icon forKey:self.callSign];
				}
				else {
					NSURLResponse *iconResponse;
					NSData *iconData = [NSURLConnection sendSynchronousRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:[NSString stringWithFormat:@"http://%@/Myth/GetChannelIcon?ChanId=%@&Height=%d", BACKEND_LOCATION, self.chanId, (int)((EPG_ROW_HEIGHT - 2) * SCALE)]]] returningResponse:&iconResponse error:NULL];
					if (iconData && (icon = [UIImage imageWithData:iconData])) {
						[response.programGuideData.channelIcons setObject:icon forKey:self.callSign];
						[iconData writeToFile:[NSString stringWithFormat:@"%@/%@.png", [documents objectAtIndex:0], self.callSign] options:0 error:NULL];
					}
					else {
						[response.programGuideData.channelIcons setObject:[NSNull null] forKey:self.callSign];
					}
				}
			}
		
			parser.delegate = response;
		}
		
		[programs sortUsingSelector:@selector(programSort:)];
		
		[pool release];
		pool = nil;
	}
}

@end
