//
//  FrontendSelect.m
//  MyMote
//
//  Copyright 2008 Chase Douglas
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  version 2 as published by the Free Software Foundation
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#import "defines.h"
#import "Backend.h"
#import "FrontendAddress.h"
#import "FrontendSelect.h"
#import "MainTabBarController.h"
#import "ProgramGuide.h"
#import "ProgramGuideView.h"
#import "RetrieveKeyBindings.h"
#import "TelnetRemote.h"

@implementation FrontendSelect

- (id)initWithBackend:(Backend *)newBackend {
	if (self = [super initWithStyle:UITableViewStylePlain]) {
		self.title = @"Frontend";
		
		backend = [newBackend retain];
	}
	return self;
}

- (void)loadView {
	frontendXML = [[NSMutableData alloc] init];
	frontends = [[NSMutableDictionary alloc] init];
	outstandingFrontendSettings = [[NSMutableArray alloc] init];	
	
	[super loadView];
	
	self.tableView.sectionFooterHeight = self.tableView.rowHeight;
	activityView = [[UIView alloc] initWithFrame:CGRectMake(0, 0, self.tableView.bounds.size.width, self.tableView.rowHeight)];
	UIActivityIndicatorView *indicator = [[[UIActivityIndicatorView alloc] initWithActivityIndicatorStyle:UIActivityIndicatorViewStyleGray] autorelease];
	indicator.frame = CGRectMake((activityView.bounds.size.width - 20) / 2, (activityView.bounds.size.height - 20) / 2, 20, 20);
	indicator.hidesWhenStopped = YES;
	indicator.tag = INDICATOR_TAG;
	[activityView addSubview:indicator];
}

- (void)viewWillAppear:(BOOL)animated {
	frontends = [[NSMutableDictionary alloc] init];
	[super viewWillAppear:animated];
	connection = [[NSURLConnection connectionWithRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:[NSString stringWithFormat:@"http://%@:%hu/Myth/GetHosts", (backend.hostname ? backend.hostname : backend.address), backend.port]]] delegate:self] retain];
	[((UIActivityIndicatorView *)[activityView viewWithTag:INDICATOR_TAG]) startAnimating];
	[self.tableView reloadData];
}

- (void)viewWillDisappear:(BOOL)animated {
	[super viewWillDisappear:animated];
	[connection cancel];
	[connection release];
	connection = nil;
	[frontends release];
	frontends = nil;
}

- (void)dealloc {
	[backend release];
	[activityView release];
	activityView = nil;
	[connection cancel];
	[connection release];
	connection = nil;
	[parser abortParsing];
	[parser release];
	parser = nil;
	[outstandingFrontendSettings release];
	[frontendXML release];
	[currentString release];
	[frontends release];
	[super dealloc];
}

// For NSURLConnection

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response {
	[frontendXML setLength:0];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data {
	[frontendXML appendData:data];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error {
	NSLog(@"Connection failed: %@ %@", [error localizedDescription], [[error userInfo] objectForKey:NSErrorFailingURLStringKey]);
	[[[[UIAlertView alloc] initWithTitle:@"Connection To Backend Failed" message:nil delegate:self cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection {
	parser = [[NSXMLParser alloc] initWithData:frontendXML];
	parser.delegate	= self;
	
	if ([parser parse] == NO) {
		[parser release];
		parser = nil;
		NSLog(@"Error Retrieving Frontends: %@", [[parser parserError] localizedDescription]);
		[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Frontends" message:nil delegate:self cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
		return;
	}
	[parser release];
	parser = nil;
}

// NSXMLParser Delegate

- (void)parser:(NSXMLParser *)parser didStartElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qualifiedName attributes:(NSDictionary *)attributeDict {
	if ([elementName isEqualToString:@"Host"]) {
		currentString  = [[NSMutableString alloc] init];
	}
}

- (void)parser:(NSXMLParser *)parser foundCharacters:(NSString *)string {
	[currentString appendString:string];
}

- (void)parser:(NSXMLParser *)parser didEndElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qualifiedName {
	if ([elementName isEqualToString:@"Host"]) {
		[outstandingFrontendSettings addObject:[[[FrontendSettings alloc] initWithBackend:backend frontend:currentString delegate:self] autorelease]];
		[currentString release];
		currentString = nil;
	}
}

// FrontendSettingsDelegate

- (void)frontendSettings:(FrontendSettings *)frontendSettings validFrontend:(NSString *)frontend port:(unsigned short)port {
	[frontends setObject:[NSNumber numberWithUnsignedShort:port] forKey:frontend];
	[self.tableView reloadData];
	
	[outstandingFrontendSettings removeObject:frontendSettings];
	if ([outstandingFrontendSettings count] <= 0) {
		[((UIActivityIndicatorView *)[activityView viewWithTag:INDICATOR_TAG]) stopAnimating];
	}
}

- (void)frontendSettings:(FrontendSettings *)frontendSettings invalidFrontend:(NSString *)frontend {
	[outstandingFrontendSettings removeObject:frontendSettings];
	if ([outstandingFrontendSettings count] <= 0) {
		[((UIActivityIndicatorView *)[activityView viewWithTag:INDICATOR_TAG]) stopAnimating];
	}
}

// UITableViewDataSource

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView {
	return 1;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section {
	return [frontends count];
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath {
	UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:@"frontendsCell"];
	if (!cell) {
		cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:@"frontendsCell"] autorelease];
	}
	
	cell.textLabel.text = [[frontends	allKeys] objectAtIndex:indexPath.row];
	cell.accessoryType = UITableViewCellAccessoryDetailDisclosureButton;
	
	return cell;
}

// UITableViewDelegate

- (UIView *)tableView:(UITableView *)tableView viewForFooterInSection:(NSInteger)section {
	return activityView;
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath {
	[DEFAULTS setBackend:backend forKey:BACKEND_KEY];
	[DEFAULTS setObject:[[frontends allKeys] objectAtIndex:indexPath.row] forKey:FRONTEND_KEY];
	[DEFAULTS setObject:[[frontends	allValues] objectAtIndex:indexPath.row] forKey:FRONTEND_PORT_KEY];

	[TELNETREMOTE setup];
	
	[PROGRAM_GUIDE loadView];
	
	[[RetrieveKeyBindings alloc] init];
	
	if (TAB_BAR_CONTROLLER.modalViewController) {
		[self.navigationController.parentViewController dismissModalViewControllerAnimated:YES];
	}
	else {
		[self.navigationController popToRootViewControllerAnimated:YES];
	}
}

- (void)tableView:(UITableView *)tableView accessoryButtonTappedForRowWithIndexPath:(NSIndexPath *)indexPath {
	UITableViewCell *frontendCell = [tableView cellForRowAtIndexPath:indexPath];
	[self.navigationController pushViewController:[[[FrontendAddress alloc] initWithBackend:[NSString stringWithFormat:@"%@:%hu", (backend.hostname ? backend.hostname : backend.address), backend.port] frontend:frontendCell.textLabel.text] autorelease] animated:YES];
}

// UIAlertViewDelegate

- (void)alertView:(UIAlertView *)alertView didDismissWithButtonIndex:(NSInteger)buttonIndex {
	[self.navigationController popViewControllerAnimated:YES];
}

@end
