//
//  ProgramsView.m
//  MyMote
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  version 2 as published by the Free Software Foundation
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#import <QuartzCore/CATiledLayer.h>

#import "Channel.h"
#import "GuideResponse.h"
#import "Layer.h"
#import "Program.h"
#import "ProgramsView.h"
#import "ProgramGuideData.h"
#import "ProgramGuideView.h"

@implementation ProgramsView

@synthesize data;

+ (Class)layerClass {
	return [CATiledLayer class];
}

- (id)init {
	if (self = [super init]) {
		self.clearsContextBeforeDrawing = NO;
		self.opaque = YES;

		layers = [[NSMutableDictionary alloc] init];
		layersLock = [[NSLock alloc] init];

		if ([self.layer respondsToSelector:@selector(contentsScale)])
			self.layer.contentsScale = SCALE;
	}
	return self;
}

- (void)setData:(ProgramGuideData *)newData {
	[data release];
	data = [newData retain];
	[self releaseLayers];
	[self setNeedsDisplay];
}

- (void)drawRect:(CGRect)rect {
	Layer *layer;

	[layersLock lock];
	if (!(layer = [layers objectForKey:[NSString stringWithFormat:@"%f,%f", rect.origin.x, rect.origin.y]])) {
		layer = [[Layer alloc] initWithCGContext:UIGraphicsGetCurrentContext() size:CGSizeMake(rect.size.width * SCALE, rect.size.height * SCALE)];
		[layers setObject:layer forKey:[NSString stringWithFormat:@"%f,%f", rect.origin.x, rect.origin.y]];
		[layer release];
		CGContextRef ctx = CGLayerGetContext(layer.layer);
		UIGraphicsPushContext(ctx);
		
		int startRow = rect.origin.y / EPG_ROW_HEIGHT;
		int endRow = startRow + rect.size.height / EPG_ROW_HEIGHT + 1;
		if (endRow >= [data.response.channels count]) {
			endRow = [data.response.channels count] - 1;
		}
		
		int row;
		for (row = startRow; row <= endRow; row++) {
			[[data.response.channels objectAtIndex:row] drawProgramsInRect:CGRectMake(rect.origin.x * SCALE, (row * EPG_ROW_HEIGHT + 1) * SCALE, rect.size.width * SCALE, (EPG_ROW_HEIGHT - 2) * SCALE) yOffset:(rect.origin.y * SCALE)];
		}
		
		UIGraphicsPopContext();
	}
	
	CGContextDrawLayerInRect(UIGraphicsGetCurrentContext(), rect, layer.layer);
	[layersLock unlock];
}

- (void)releaseLayers {
	[layersLock lock];
	[layers release];
	layers = [[NSMutableDictionary alloc] init];
	[layersLock unlock];
}

- (Program *)programAtPoint:(CGPoint)point {
	Channel *channel = [data.response.channels objectAtIndex:((self.frame.origin.y + point.y) / EPG_ROW_HEIGHT)];
	
	NSDateComponents *offset = [[NSDateComponents alloc] init];
	offset.second = data.response.guideMinuteLength * 60 * (self.frame.origin.x + point.x) / EPG_PROGRAMS_WIDTH;
	NSDate *offsetDate = [[NSCalendar currentCalendar] dateByAddingComponents:offset toDate:data.response.startDate options:0];
	[offset release];
	
	for (Program *program in channel.programs) {
		if ([program.startDate compare:offsetDate] != NSOrderedDescending && [program.endDate compare:offsetDate] != NSOrderedAscending) {
			return program;
		}
	}
	
	return nil;
}

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event {
	if ([touches count] == 1) {
		touchedProgram = [self programAtPoint:[[touches anyObject] locationInView:self]];
	}
	else {
		touchedProgram = nil;
	}
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event {
	if ([touches count] == 1 && [self pointInside:[[touches anyObject] locationInView:self] withEvent:nil]) {
		if (touchedProgram && touchedProgram == [self programAtPoint:[[touches anyObject] locationInView:self]]) {
			[(ProgramGuideView *)self.superview.superview showProgramInfo:touchedProgram];
		}
	}
}

- (void)dealloc {
	[layersLock lock];
	[layers release];
	layers = nil;
	[layersLock unlock];
	[layersLock release];
	[data release];
	[super dealloc];
}

@end
