//
//  RetrieveKeyBindings.m
//  MyMote
//
//  Copyright 2008 Chase Douglas
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  version 2 as published by the Free Software Foundation
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#import <mysql/mysql.h>

#import "RetrieveKeyBindings.h"

#define NO_HOST 0
#define HAS_HOST 1

@implementation RetrieveKeyBindings

- (id)init {
	if (self = [super init]) {
		connectionInfo = [[NSMutableDictionary alloc] init];
		currentString = [[NSMutableString alloc] init];
		
		lock = [[NSLock alloc] init];
		
		[DEFAULTS setObject:[[[NSDictionary alloc] initWithContentsOfFile:[[NSBundle mainBundle] pathForResource:@"Keys" ofType:@"plist"]] autorelease] forKey:@"Keys"];
		
		xml = [[NSMutableData alloc] init];
		connection = [[NSURLConnection connectionWithRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:[NSString stringWithFormat:@"http://%@/Myth/GetConnectionInfo?Pin=%@", BACKEND_LOCATION, [[DEFAULTS objectForKey:BACKEND_PIN_KEY] objectForKey:BACKEND_LOCATION]]]] delegate:self] retain];

		retrieveAlertView = [[UIAlertView alloc] initWithTitle:@"Retrieving Key Bindings" message:@"If you cancel, the MythTV default bindings will be used. Remote buttons may not function correctly" delegate:self cancelButtonTitle:nil otherButtonTitles:@"Cancel", nil];
		[retrieveAlertView show];
	}
	return self;
}

- (void)dealloc {
	[retrieveAlertView release];
	[connectionInfo release];
	[currentString release];
	[connection release];
	[xml release];
	[lock release];
	[super dealloc];
}

// For NSURLConnection

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response {
	[xml setLength:0];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data {
	[xml appendData:data];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error {
	[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
	NSLog(@"GetConnectionInfo failed: %@ %@", [error localizedDescription], [[error userInfo] objectForKey:NSErrorFailingURLStringKey]);
	[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Key Bindings" message:@"Mythtv default bindings will be used. Remote buttons may not function correctly." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection {
	[lock lock];
	
	if (canceled) {
		[lock unlock];
		return;
	}
	
	parser = [[NSXMLParser alloc] initWithData:xml];
	parser.delegate	= self;
	
	if ([parser parse] == NO) {
		[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
		[lock unlock];
		if (!canceled) {
			[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Key Bindings" message:@"Mythtv default bindings will be used. Remote buttons may not function correctly." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
			NSLog(@"Could not parse GetConnectionInfo");
		}
		[parser release];
		return;
	}
	[parser release];
	
	if ([[connectionInfo objectForKey:@"errorCode"] isEqualToString:@"604"]) {
		[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
		[lock unlock];
		[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Key Bindings" message:@"The MythTV backend does not have the security pin set. Run mythtv-setup to set the pin.\n\nMythtv default bindings will be used. Remote buttons may not function correctly." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
		NSLog(@"Received error code 604 from GetConnectionInfo");
		return;
	}
	
	if ([[connectionInfo objectForKey:@"errorCode"] isEqualToString:@"606"]) {
		[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
		[lock unlock];
		[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Key Bindings" message:@"The backend security pin was not valid. Please enter the correct security pin when selecting the backend.\n\nMythtv default bindings will be used. Remote buttons may not function correctly." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
		NSLog(@"Received error code 606 from GetConnectionInfo");
		return;
	}
	
	if ([connectionInfo objectForKey:@"errorCode"]) {
		NSLog(@"Received error code %@ from GetConnectionInfo", [connectionInfo objectForKey:@"errorCode"]);
	}
	
	MYSQL *mysql = mysql_init(NULL);
	if (!mysql) {
		[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
		[lock unlock];
		[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Key Bindings" message:@"Mythtv default bindings will be used. Remote buttons may not function correctly." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
		NSLog(@"Error while initializing MySQL: (%d) %s", mysql_errno(mysql), mysql_error(mysql));
		return;
	}
	
	if ([[DEFAULTS objectForKey:DATABASE_ADDRESS_KEY] objectForKey:BACKEND_LOCATION]) {
		host = [[DEFAULTS objectForKey:DATABASE_ADDRESS_KEY] objectForKey:BACKEND_LOCATION];
	}
	else {
		host = [connectionInfo objectForKey:@"Host"];
	}
	
	if (!mysql_real_connect(mysql, [host UTF8String], [[connectionInfo objectForKey:@"UserName"] UTF8String], [[connectionInfo objectForKey:@"Password"] UTF8String], [[connectionInfo objectForKey:@"Name"] UTF8String], [[connectionInfo objectForKey:@"Port"] intValue], NULL, CLIENT_COMPRESS)) {
		[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
		[lock unlock];
		[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Key Bindings" message:@"Mythtv default bindings will be used. Remote buttons may not function correctly." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
		NSLog(@"Error while connecting to MySQL database at host '%@': (%d) %s", host, mysql_errno(mysql), mysql_error(mysql));
		mysql_close(mysql);
		return;
	}
	
	thread_id = mysql->thread_id;
	[lock unlock];
	
	if (mysql_query(mysql, [[NSString stringWithFormat:@"SELECT context, action, keylist FROM keybindings WHERE hostname = \"%@\"", FRONTEND] UTF8String])) {
		[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
		if (!canceled) {
			[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Key Bindings" message:@"Mythtv default bindings will be used. Remote buttons may not function correctly." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
			NSLog(@"Error while querying MySQL Server for key bindings: (%d) %s", mysql_errno(mysql), mysql_error(mysql));
		}
		mysql_close(mysql);
		return;
	}
	
	MYSQL_RES *result = mysql_store_result(mysql);
	if (!result) {
		[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
		if (!canceled) {
			[[[[UIAlertView alloc] initWithTitle:@"Could Not Retrieve Key Bindings" message:@"Mythtv default bindings will be used. Remote buttons may not function correctly." delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
			NSLog(@"Error while storing key bindings query result: (%d) %s", mysql_errno(mysql), mysql_error(mysql));
		}
		mysql_close(mysql);
		return;
	}
	
	NSMutableDictionary *keys = [[NSMutableDictionary alloc] init];
	
	MYSQL_ROW row;
	while (!canceled && (row = mysql_fetch_row(result))) {
		NSString *context = [NSString stringWithCString:row[0] encoding:NSUTF8StringEncoding];
		if (![keys objectForKey:context]) {
			[keys setObject:[NSMutableDictionary dictionary] forKey:context];
		}
		
		NSString *key;
		if (!row[2]) {
			continue;
		}
		else if (strlen(row[2]) > 0 && row[2][0] == ',') {
			key = [@"," retain];
		}
		else {
			NSArray *possibleKeys = [[NSString stringWithCString:row[2] encoding:NSUTF8StringEncoding] componentsSeparatedByCharactersInSet:[NSCharacterSet characterSetWithCharactersInString:@","]];
			key = [[[possibleKeys objectAtIndex:0] lowercaseString] retain];
		}
		
		if ([key isEqualToString:@"esc"]) {
			[key release];
			key = [@"escape" retain];
		}
		else if ([key isEqualToString:@"pgup"]) {
			[key release];
			key = [@"pageup" retain];
		}
		else if ([key isEqualToString:@"pgdown"]) {
			[key release];
			key = [@"pagedown" retain];
		}
		else if ([key isEqualToString:@"return"]) {
			[key release];
			key = [@"enter" retain];
		}
		
		NSString *action = [NSString stringWithCString:row[1] encoding:NSUTF8StringEncoding];
		[[keys objectForKey:context] setObject:[NSDictionary dictionaryWithObjectsAndKeys:key, @"Key", nil] forKey:action];
		
		[key release];
	}
	
	mysql_free_result(result);
	mysql_close(mysql);
	
	[DEFAULTS setObject:keys forKey:@"Keys"];
	
	[keys release];
	
	[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
}

// NSXMLParser Delegate

- (void)parser:(NSXMLParser *)parser didStartElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qualifiedName attributes:(NSDictionary *)attributeDict {
	[currentString setString:@""];
}

- (void)parser:(NSXMLParser *)parser foundCharacters:(NSString *)string {
	[currentString appendString:string];
}

- (void)parser:(NSXMLParser *)parser didEndElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qualifiedName {
	if ([elementName isEqualToString:@"Host"] || [elementName isEqualToString:@"Port"] || [elementName isEqualToString:@"UserName"] || [elementName isEqualToString:@"Password"] || [elementName isEqualToString:@"Name"] || [elementName isEqualToString:@"errorCode"]) {
		[connectionInfo setObject:[NSString stringWithString:currentString] forKey:elementName];
	}
}

// UIAlertViewDelegate

- (void)alertView:(UIAlertView *)alertView clickedButtonAtIndex:(NSInteger)buttonIndex {
	canceled = YES;
	
	[connection cancel];
	[parser abortParsing];
	
	[lock lock];
	[lock unlock];
	
	if (!thread_id) {
		[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
		return;
	}
	
	MYSQL *killMysql = mysql_init(NULL);
	if (!killMysql) {
		[[[[UIAlertView alloc] initWithTitle:@"Could Not Cancel Key Bindings Retrieval" message:nil delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
		return;
	}
	
	if (!host) {
		return;
	}
	
	if (!mysql_real_connect(killMysql, [host UTF8String], [[connectionInfo objectForKey:@"UserName"] UTF8String], [[connectionInfo objectForKey:@"Password"] UTF8String], [[connectionInfo objectForKey:@"Name"] UTF8String], [[connectionInfo objectForKey:@"Port"] intValue], NULL, CLIENT_COMPRESS)) {
		[[[[UIAlertView alloc] initWithTitle:@"Could Not Cancel Key Bindings Retrieval" message:nil delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
		NSLog(@"killMysql Connection Error: (%d) %s", mysql_errno(killMysql), mysql_error(killMysql));
		mysql_close(killMysql);
		return;
	}
	
	if (mysql_query(killMysql, [[NSString stringWithFormat:@"KILL QUERY %lu", thread_id] UTF8String])) {
		if (mysql_errno(killMysql) != 1094) { // 1094 means there was no thread, hence query is already complete
			[[[[UIAlertView alloc] initWithTitle:@"Could Not Cancel Key Bindings Retrieval" message:nil delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil] autorelease] show];
			mysql_close(killMysql);
			return;
		}
	}
	
	[retrieveAlertView dismissWithClickedButtonIndex:0 animated:YES];
			
	MYSQL_RES *result = mysql_store_result(killMysql);
	if (result) {
		mysql_free_result(result);
	}
	
	mysql_close(killMysql);
}

@end
