//
//  crashreporter.c
//  MyMote
//
//  Copyright 2008 Chase Douglas
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  version 2 as published by the Free Software Foundation
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#import <dlfcn.h>
#import <execinfo.h>
#import <objc/runtime.h>
#import <signal.h>
#import <ucontext.h>

#import "version.h"
#include "crashreporter.h"

/*
 * This function takes a report to be e-mailed and creates a mailto: URL that the
 * UIApplication object can handle. The report is reformatted to conform to the URL
 * standards (replacing ' ' with '%20' for example). This is appended to a
 * preformed message header with the To: and Subject: fields already populated.
 * Note that the recipient is obfuscated by being split into multiple lines to
 * prevent spam.
 *
 * This functionality is gated by the "Send Crash Reports" user preferences
 * setting.
 */
static void emailReport(NSString *report) {
	if ([[NSUserDefaults standardUserDefaults] boolForKey:@"Send Crash Reports"]) {
		NSMutableCharacterSet *validURLChars = [NSMutableCharacterSet alphanumericCharacterSet];
		[validURLChars addCharactersInString:@"$-_@.&+-!*\"'|,"];
		
		NSMutableString *url = [NSMutableString stringWithString:@"mailto:mymote"];
		[url appendString:@".crash"];
		[url appendString:@".reports"];
		[url appendString:@"@gmail.com?subject=MyMote%20Crash%20Report&body="];
		NSScanner *scanner = [NSScanner scannerWithString:report];
		while (![scanner isAtEnd]) {
			NSString *temp;
			[scanner scanCharactersFromSet:validURLChars intoString:&temp];
			if (temp) {
				[url appendString:temp];
			}
			if (![scanner isAtEnd]) {
				uint8_t c = [report characterAtIndex:[scanner scanLocation]];
				while (![validURLChars characterIsMember:c]) {
					[url appendFormat:@"%%%02hhx", c];
					[scanner setScanLocation:([scanner scanLocation] + 1)];
					if (![scanner isAtEnd]) {
						c = [report characterAtIndex:[scanner scanLocation]];
					}
					else {
						break;
					}
				}
			}
		}
		
		[[UIApplication sharedApplication] openURL:[NSURL URLWithString:url]];
	}
}

#if __IPHONE_OS_VERSION_MIN_REQUIRED < __IPHONE_3_0
// This may be used for iPhone OS < 3.0. iPhone OS 3.0 and greater already provide this functionality.
static NSString *backtraceFromFrames(void *frames[], int numFrames) {
	int numClasses = objc_getClassList(NULL, 0);
	Class classes[numClasses];
	numClasses = objc_getClassList(classes, numClasses);
	
	Method *classMethods[numClasses];
	unsigned int classMethodCounts[numClasses];
	
	int i;
	for (i = 0; i < numClasses; i++) {
		classMethods[i] = class_copyMethodList(object_getClass(classes[i]), &classMethodCounts[i]);
	}
	
	Method *instanceMethods[numClasses];
	unsigned int instanceMethodCounts[numClasses];
	
	for (i = 0; i < numClasses; i++) {
		instanceMethods[i] = class_copyMethodList(classes[i], &instanceMethodCounts[i]);
	}
	
	NSMutableString *trace = [[NSMutableString alloc] init];
	
	for (i = 0; i < numFrames; i++) {
		BOOL classMethod;
		const char *className;
		const char *methodName;
		void *functionAddress = NULL;
		Dl_info dlInfo;
		BOOL dlFunction = YES;
		
		if (dladdr(frames[i], &dlInfo)) {
			if (dlInfo.dli_saddr) {
				functionAddress = dlInfo.dli_saddr;
			}
			else {
				functionAddress = dlInfo.dli_fbase;
			}
		}
		
		int j;
		for (j = 0; j < numClasses; j++) {
			int k;
			for (k = 0; k < classMethodCounts[j]; k++) {
				void *address = method_getImplementation(classMethods[j][k]);
				if (address < frames[i] && address > functionAddress) {
					functionAddress = address;
					methodName = sel_getName(method_getName(classMethods[j][k]));
					className = class_getName(classes[j]);
					classMethod = YES;
					dlFunction = NO;
				}
			}
			
			for (k = 0; k < instanceMethodCounts[j]; k++) {
				void *address = method_getImplementation(instanceMethods[j][k]);
				if (address < frames[i] && address > functionAddress) {
					functionAddress = address;
					methodName = sel_getName(method_getName(instanceMethods[j][k]));
					className = class_getName(classes[j]);
					classMethod = NO;
					dlFunction = NO;
				}				
			}
		}
		
		if (dlFunction) {
			if (dlInfo.dli_sname) {
				[trace appendFormat:@"%d %p %s %s + %d\n", i, frames[i], dlInfo.dli_fname, dlInfo.dli_sname, frames[i] - functionAddress];
			}
			else {
				[trace appendFormat:@"%d %p %s + %d\n", i, frames[i], dlInfo.dli_fname, frames[i] - functionAddress];
			}
		}
		else {
			if (classMethod) {				
				[trace appendFormat:@"%d %p +[%s %s] + %d\n", i, frames[i], className, methodName, frames[i] - functionAddress];
			}
			else {
				[trace appendFormat:@"%d %p -[%s %s] + %d\n", i, frames[i], className, methodName, frames[i] - functionAddress];
			}
		}
	}
	
	return trace;
}
#endif

/*
 * This function formulates a crash report with data available from the NSException
 * object. The most useful pieces of data are the exception reason and the
 * backtrace. The backtrace addresses are formulated into a more readable format.
 * The report is then logged to stderr and then emailed to a central address.
 */
static void uncaughtExceptionHandler(NSException *exception) {
	// Transpose the NSException backtrace NSArray object into an array of void
	// pointers for backtrace_symbols.
	void *backtraceFrames[[[exception callStackReturnAddresses] count]];
	int i = 0;
	for (NSNumber *frame in [exception callStackReturnAddresses]) {
		backtraceFrames[i++] = (void *)[frame unsignedIntegerValue];
	}
	int frameCount = [[exception callStackReturnAddresses] count];
	
	// Get symbols for the backtrace addresses
	char **frameStrings = backtrace_symbols(backtraceFrames, frameCount);
	
	NSMutableString *report = [NSMutableString string];
	
	// Populate header of report with exception information
	[report appendFormat:@"Unhandled %@ exception encountered\n", [exception name]];
	[report appendFormat:@"Reason: %@\n", [exception reason]];
	[report appendString:@"MyMote Revision: " MYMOTE_SVN_REVISION @"\n\n"];
	if ([[exception userInfo] count]) {
		[report appendString:@"Application Data:\n"];
		[report appendString:[[exception userInfo] description]];
		[report appendString:@"\n\n"];
	}
	else {
		[report appendString:@"No Application Data Available\n\n"];
	}
	
	// Start backtrace
	[report appendString:@"Backtrace:"];
	
	if(frameStrings != NULL) {
		int i = 0;
		
		// Append each stack frame string to the report
		for(i = 0; i < frameCount; i++) {
			if(frameStrings[i] == NULL) {
				continue;
			}
			[report appendFormat:@"\n%s", frameStrings[i]];
		}
	}
	
	// Log the report to stderr
	NSLog(report);
	
	// Send the report out
	emailReport(report);
	
	// This is needed because we're about to end the program
	[[NSUserDefaults standardUserDefaults] synchronize];
}

/*
 * This function handles signals like segmentation faults. Signal details and a
 * stack trace are reported. Note that this function retrieves the stack by peering
 * into the registers of the process, and some inferences may become inaccurate if
 * the underlying stack frame structures are modified in the future.
 */
static void mysighandler(int sig, siginfo_t *info, void *context) {
	NSMutableString *report = [NSMutableString string];
	[report appendFormat:@"MyMote terminated due to a %s signal\n", sys_signame[sig]];
	[report appendString:@"MyMote Revision: " MYMOTE_SVN_REVISION @"\n"];
	[report appendFormat:@"Signal Number: %d\n", info->si_signo];
	[report appendFormat:@"Error Number: %d\n", info->si_errno];
	[report appendFormat:@"Signal Code: %d\n", info->si_code];
	[report appendFormat:@"PID: %d\n", info->si_pid];
	[report appendFormat:@"UID: %d\n", info->si_uid];
	[report appendFormat:@"Exit Value: %d\n", info->si_status];
	[report appendFormat:@"Address: %p\n", info->si_addr];
	[report appendFormat:@"Signal Value: %p\n", info->si_value];
	[report appendFormat:@"Band: %ld\n", info->si_band];
	
	[report appendString:@"\nBacktrace:\n"];
	
	ucontext_t *ucontext = context;
	
	// Prepare backtrace frames array
	void *backtraceFrames[128];
	int frameCount = 0;
	
	/* These blocks perform stack unwinding to get the backtrace. The registers and
	 * stack frame layouts are architecture specific. Note: ip is the instruction
	 * pointer, bp is the base pointer, and fp is the frame pointer.
	 */
#ifdef __i386__
	void *ip = (void *)ucontext->uc_mcontext->__ss.__eip;
	void **bp = (void **)ucontext->uc_mcontext->__ss.__ebp;
	while (ip && bp && frameCount < 128) {
		backtraceFrames[frameCount++] = ip;
		ip = bp[1];
		bp = bp[0];
	}
#elif defined(__arm__) && defined(__thumb__)
	void *ip = (void *)ucontext->uc_mcontext->__ss.__pc;
	void **fp = (void *)ucontext->uc_mcontext->__ss.__r[7]; // General purpose register r7 is the frame pointer register for the ARM thumb architecture
	void **prevfp = fp - 1;
	
	// Check for word alignment and make sure frames have increasing addresses
	while (ip && fp && !((uint32_t)fp & 3) && prevfp < fp && frameCount < 128) {
		backtraceFrames[frameCount++] = ip;
		// These locations seem to be incorrect according to the ARM docs I've
		// seen online, but they seem to work on the iphone.
		ip = fp[1];
		prevfp = fp;
		fp = fp[0];
	}
#else
	// No support for other architectures
	backtraceFrames[0] = NULL;
	[report appendString:@"No backtrace available on this architecture"];
#endif
	
	// Convert the backtrace addresses into readable strings
	char **frameStrings = backtrace_symbols(backtraceFrames, frameCount);
	
	if(frameStrings != NULL) {
		int i = 0;
		
		// Append each stack frame string to the report
		for(i = 0; i < frameCount; i++) {
			if(frameStrings[i] == NULL) {
				continue;
			}
			[report appendFormat:@"\n%s", frameStrings[i]];
		}
	}
	
#if defined(__arm__) && defined(__thumb__)
	// Check if the backtrace was aborted due to frame word alignment or
	// stack growth direction. The last instruction pointer should be NULL
	// unless the backtrace was aborted, so check it first.
	if (ip) {
		if ((uint32_t)fp & 3) {
			[report appendFormat:@"\nError: Next frame pointer is not word aligned (%p)", fp];
		}
		else if (fp && prevfp >= fp) {
			[report appendFormat:@"\nError: Next frame pointer is up the stack (Next: %p, Previous: %p)", fp, prevfp];
		}
	}
#endif
	
	// Log the report to stderr
	NSLog(report);
	
	// Send the report out
	emailReport(report);
	
	// This is needed because we're about to end the program
	[[NSUserDefaults standardUserDefaults] synchronize];
	
	exit(-1);
}

int crashReporterMain(int argc, char *argv[], NSString *applicationDelegateName) {
	// Install the signal handler for the following signals
	struct sigaction newsigaction;
	newsigaction.sa_sigaction = mysighandler;
	newsigaction.sa_flags = SA_SIGINFO;
	sigemptyset(&newsigaction.sa_mask);
	sigaction(SIGQUIT, &newsigaction, NULL);
	sigaction(SIGILL, &newsigaction, NULL);
	sigaction(SIGTRAP, &newsigaction, NULL);
	sigaction(SIGABRT, &newsigaction, NULL);
	sigaction(SIGEMT, &newsigaction, NULL);
	sigaction(SIGFPE, &newsigaction, NULL);
	sigaction(SIGBUS, &newsigaction, NULL);
	sigaction(SIGSEGV, &newsigaction, NULL);
	sigaction(SIGSYS, &newsigaction, NULL);
	sigaction(SIGPIPE, &newsigaction, NULL);
	sigaction(SIGALRM, &newsigaction, NULL);
	sigaction(SIGXCPU, &newsigaction, NULL);
	sigaction(SIGXFSZ, &newsigaction, NULL);
	
	int retVal;
	
	// The following exception handling logic allows us to catch unhandled
	// exceptions. Using the builtin unhandled exceptions handler routines
	// fails to allow the handler to open the mailto: URL for sending crash
	// reports.
	@try {
		retVal = UIApplicationMain(argc, argv, nil, applicationDelegateName);
	}
	@catch (NSException *exception) {
		uncaughtExceptionHandler(exception);
		retVal = EXIT_FAILURE;
	}
	
    return retVal;
}
